package com.redlimerl.speedrunigt.timer.packet.packets;

import com.redlimerl.speedrunigt.SpeedRunIGT;
import com.redlimerl.speedrunigt.option.SpeedRunOption;
import com.redlimerl.speedrunigt.option.SpeedRunOptions;
import com.redlimerl.speedrunigt.timer.InGameTimer;
import com.redlimerl.speedrunigt.timer.category.RunCategory;
import com.redlimerl.speedrunigt.timer.packet.TimerPacket;
import com.redlimerl.speedrunigt.timer.packet.TimerPacketBuf;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.client.MinecraftClient;
import net.minecraft.server.MinecraftServer;
import net.minecraft.util.Identifier;

public class TimerChangeCategoryPacket extends TimerPacket {

    public static final Identifier IDENTIFIER = TimerPacket.identifier("timer_category");
    private final String sendCategory;

    public TimerChangeCategoryPacket() {
        this(null);
    }

    public TimerChangeCategoryPacket(String name) {
        super(IDENTIFIER);
        this.sendCategory = name;
    }

    @Environment(EnvType.CLIENT)
    @Override
    protected TimerPacketBuf convertClient2ServerPacket(TimerPacketBuf buf, MinecraftClient client) {
        if (sendCategory != null) {
            buf.writeString(sendCategory);
        }
        return buf;
    }

    @Override
    public void receiveClient2ServerPacket(TimerPacketBuf buf, MinecraftServer server) {
        if (!SpeedRunIGT.IS_CLIENT_SIDE) {
            TimerPacketBuf copiedBuf = buf.copy();
            RunCategory runCategory = RunCategory.getCategory(copiedBuf.readString());
            InGameTimer.getInstance().setCategory(runCategory, false);
            SpeedRunOption.setOption(SpeedRunOptions.TIMER_CATEGORY, runCategory);
            copiedBuf.release();
        }
        this.sendPacketToPlayers(buf, server);
    }

    @Override
    protected TimerPacketBuf convertServer2ClientPacket(TimerPacketBuf buf, MinecraftServer server) {
        if (sendCategory != null) {
            buf.writeString(sendCategory);
        }
        return buf;
    }

    @Environment(EnvType.CLIENT)
    @Override
    public void receiveServer2ClientPacket(TimerPacketBuf buf, MinecraftClient client) {
        InGameTimer.getInstance().setCategory(RunCategory.getCategory(buf.readString()), false);
    }
}
